<?

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
    *                                                                 *
    *   BROADCAST Chat Server 1.3                                     *
    *   Copyright  2004 Angelo Cuccato <angelo.cuccato@epost.de>     *
    *                    Sebastian Pipping <webmaster@hartwork.org>   *
    *                                                                 *
    *   --> http://www.hartwork.org                                   *
    *                                                                 *
    *                                                                 *
    *   This source code is released under LGPL.                      *
    *   See LGPL.txt for details.                        2004-09-15   *
    *                                                                 *
    *   ___________________________________________________________   *
    *                                                                 *
    *   Big thanx to Ori Staub for wrting the tutorial                *
    *   "Writing Socket Servers in PHP" which can be found on         *
    *   http://www.zend.com/zend/tut/tutorial-staub3.php              *
    *                                                                 *
    \* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


// (1) Uncomment "extension=php_sockets.dll" in php.ini
// (2) Make sure you edit the right php.ini if you use PHP with Apache
// (3) Make sure your firewall is not blocking us


// Settings
define( SERVER_ADDRESS,	"0.0.0.0"	);
define( SERVER_PORT,	9000		);
define( MAX_CLIENTS,	50			);

define( BCS_VERSION,	"1.3 // www.hartwork.org" );
define( CRLF,			"\015\012"	);


// Get unlimited runtime
set_time_limit( 0 );

// Turn off buffering
ob_end_clean();


// Clients
$clients = array();
$active_clients = 0;


// Set us up as server
$sock = socket_create( AF_INET, SOCK_STREAM, 0 );
socket_bind( $sock, SERVER_ADDRESS, SERVER_PORT ) or die( "Could not bind " . SERVER_ADDRESS . ":" . SERVER_PORT );
socket_listen( $sock );

echo "Running\n";

while( TRUE )
{
	$read = array();

	// Copy clients to [$read]
	$read[ 0 ] = $sock;
	$i = 0;
	foreach( $clients as $val )
	{
		$read[ ++$i ] = $val[ "sock" ];
	}

	$ready = socket_select( $read, $write, $except, 0 );

	// Handle new clients
	if( in_array( $sock, $read ) )
	{
		BCS_add_client( $sock );
	}

	// Serve clients
	for( $i = 0; $i < $active_clients; $i++ )
	{
		if( in_array( BCS_get_sock( $i ), $read ) )
		{
			// Append input
			$temp = BCS_read( $i );
			if( is_bool( $temp ) )
			{
				BCS_kick_client( $i );
				continue;
			}

			$input = BCS_get_input( $i ) . $temp;

			// Apply backspaces
			$input = preg_replace( "/(^|.)\\x08/", "", $input );

			// IF END OF STRING = <crlf>
			if( preg_match( "/[\\015\\012]{2}$/", $input ) )
			{
				// Kick <crlf> and trailing spaces
				$input = preg_replace( "/[ \\015\\012]+$/", "", $input );


				if( $input == "" )
				{
					BCS_reset_input( $i );
				}
				else
				{
					// Evaluate input
					BCS_set_input( $i, $input );
					BCS_eval( $i );
				}
			}
			else
			{
				BCS_set_input( $i, $input );
			}
		}
	}
}

// Close master sockets
socket_close( $sock );


/****************************************************************************************************/
//****** BCS_add_client
/****************************************************************************************************/
function BCS_add_client( $sock )
{
	global $clients, $active_clients;
	if( $active_clients < MAX_CLIENTS )
	{
		$clients[ $active_clients ] = array();

		BCS_set_sock( $active_clients, socket_accept( $sock ) );
		BCS_reset_input( $active_clients );

		socket_getpeername( BCS_get_sock( $active_clients ), $ip, $port );
		BCS_set_addr( $active_clients, "$ip:$port" );

		// Say hello
		BCS_write( $active_clients, "BROADCAST Chat Server " . BCS_VERSION . CRLF );
		BCS_socket_set_nonblock( $active_clients );

		// Server log
		echo "[" . BCS_get_addr( $active_clients ) . "] connected\n";

		$active_clients++;
	}
}

/****************************************************************************************************/
//****** BCS_kick_client
/****************************************************************************************************/
function BCS_kick_client( $index )
{
	global $clients, $active_clients;
	
	echo "[" . BCS_get_addr( $index ) . "] disconnected\n";
	
	@socket_close( $clients[ $index ][ "sock" ] );
	$clients[ $index ] = $clients[ --$active_clients ];
	array_pop( $clients );
}


/****************************************************************************************************/
//****** BCS_eval
/****************************************************************************************************/
function BCS_eval( $index )
{
	global $clients, $active_clients;
	$allow_reset = TRUE;
	$input = BCS_get_input( $index );
	
	if( strncmp( $input, "#", 1 ) )
	{
		// No command -> message
		if( BCS_get_nick( $index ) != "" )
		{
			// Has nick -> broadcast
			for( $i = 0; $i <= $active_clients; $i++ )
			{
				if( BCS_get_nick( $i ) != "" )		// To logged-in only
				{
					BCS_write( $i, date( "Y-d-m H:i:s" ) . " [" . BCS_get_nick( $index ) . "]: " . BCS_get_input( $index ) . CRLF );
				}
			}
		}
		else
		{
			// No nick -> error
			BCS_write( $index, "ERROR: You are not logged in. Use [#nick <nickname>] to join the chat." . CRLF );
		}

		// Reset buffer
		BCS_reset_input( $index );
	}
	else
	{
		// Command
		
		$params  = "";
		$matches = array();


/****************************************************************************************************/
//***** #nick <nickname>
/****************************************************************************************************/

		if( !strncasecmp( $input, "#nick", 5 ) )
		{
			$params = substr( $input, 5 );
			if( ( $params == "" ) || preg_match( "/^ +$/", $params ) )
			{
				// Wrong syntax
				BCS_write( $index, "ERROR: Parameter missing" . CRLF );
			}
			else if( preg_match( "/^ +([^ ]+)$/", $params, $matches ) )
			{
				$nick_old = BCS_get_nick( $index );
				$nick_new = preg_replace( "/[,\\[\\]]/", "", $matches[ 1 ] );

				if( $nick_new == "" )
				{
					BCS_write( $index, "ERROR: Invalid nickname" . CRLF );
				}
				else
				{
					// Nick already used?
					$found = FALSE;
					for( $i = 0; $i <= $active_clients; $i++ )
					{
						if( $i == $index ) continue;
						if( !strcasecmp( BCS_get_nick( $i ), $nick_new ) )
						{
							$found = TRUE;
							break; // Stop searching
						}
					}

					if( $found )
					{
						// Nick already used
						BCS_write( $index, "ERROR: Nickname already in use" . CRLF );
					}
					else
					{
						BCS_set_nick( $index, $nick_new );

						// Notify others
						if( $nick_old == "" )
						{
							// User has just logged in
							for( $i = 0; $i <= $active_clients; $i++ )
							{
								if( $i == $index ) continue;
								if( BCS_get_nick( $i ) != "" )	// To logged-in only
								{
									BCS_write( $i, "[" . $nick_new . "] joined the chat." . CRLF );
								}
							}

							BCS_write( $index, "You joined the chat as [" . $nick_new . "]." . CRLF );

							echo "[" . BCS_get_addr( $index ) . "] is [" . $nick_new . "]\n";
						}
						else
						{
							// User changed its nick
							for( $i = 0; $i <= $active_clients; $i++ )
							{
								if( $i == $index ) continue;
								if( BCS_get_nick( $i ) != "" )	// To logged-in only
								{
									BCS_write( $i, "[" . $nick_old . "] is now [" . $nick_new . "]" . CRLF );
								}
							}

							BCS_write( $index, "Your nick was changed to [" . $nick_new . "]" . CRLF );

							echo "[" . BCS_get_addr( $index ) . "] is now [" . $nick_new . "]\n";
						}
					}
				}
			}
			else
			{
				// Wrong syntax
				BCS_write( $index, "ERROR: A nickname must NOT contain spaces." . CRLF );
			}
		}


/****************************************************************************************************/
//***** #list
/****************************************************************************************************/

		else if( !strncasecmp( $input, "#list", 5 ) )
		{
			$params = substr( $input, 5 );
			if( ( $params == "" ) || preg_match( "/^ +$/", $params ) )
			{
				if( BCS_get_nick( $index ) == "" )
				{
					BCS_write( $index, "ERROR: You are not logged in. Use [#nick <nickname>] to join the chat." . CRLF );
				}
				else
				{
					// List users
					$list	= "";
					$users	= 0;

					for( $i = 0; $i <= $active_clients; $i++ )
					{
						$nick = BCS_get_nick( $i );
						if( $nick != "" )
						{
							$list .= ( ( $users++ > 0 ) ? ", " : "" ) . "[" . $nick . "]";
						}
					}

					if( $list != "" )
					{
						BCS_write( $index, "$users user" . ( ( $users == 1 ) ? "" : "s" ) . ": $list" . CRLF );
					}
				}
			}
			else
			{
				// Wrong syntax
				BCS_write( $index, "ERROR: [#list] takes no parameters." . CRLF );
			}
		}


/****************************************************************************************************/
//***** #quit
/****************************************************************************************************/

		else if( !strncasecmp( $input, "#quit", 5 ) )
		{
			$params = substr( $input, 5 );
			if( ( $params == "" ) || preg_match( "/^ +$/", $params ) )
			{
				// Notify others
				for( $i = 0; $i <= $active_clients; $i++ )
				{
					if( $i == $index ) continue;
					if( BCS_get_nick( $i ) != "" )		// To logged-in only
					{
						BCS_write( $i, "[" . BCS_get_nick( $index ) . "] left the chat" . CRLF );
					}
				}

				// Say goodbye
				BCS_write( $index, "You left the chat" . CRLF );

				// Kick client
				BCS_kick_client( $index );

				// DO NOT RESET BUFFER !!!
				$allow_reset = FALSE;
			}
			else
			{
				// Wrong syntax
				BCS_write( $index, "ERROR: [#quit] takes no parameters." . CRLF );
			}
		}

/****************************************************************************************************/
//***** #...
/****************************************************************************************************/

		else
		{
			BCS_write( $index, "ERROR: Unknown command" . CRLF );
		}

		// Reset buffer
		if( $allow_reset ) BCS_reset_input( $index );
	}
}

/****************************************************************************************************/
//****** BCS_read
/****************************************************************************************************/
function BCS_read( $index )
{
	global $clients;
	if( is_bool( @socket_recv( $clients[ $index ][ "sock" ], $buf, 1024, 0 ) ) )
	{
		return FALSE;
	}
	else
	{
		return $buf;
	}
}

/****************************************************************************************************/
//****** BCS_write
/****************************************************************************************************/
function BCS_write( $index, $text )
{
	global $clients;
	@socket_send( $clients[ $index ][ "sock" ], $text, strlen( $text ), 0 );
}

/****************************************************************************************************/
//****** BCS_socket_set_nonblock
/****************************************************************************************************/
function BCS_socket_set_nonblock( $index )
{
	global $clients;
	socket_set_nonblock( $clients[ $index ][ "sock" ] );
}

/****************************************************************************************************/
//****** BCS_get_sock
/****************************************************************************************************/
function BCS_get_sock( $index )
{
	global $clients;
	return $clients[ $index ][ "sock" ];
}

/****************************************************************************************************/
//****** BCS_set_sock
/****************************************************************************************************/
function BCS_set_sock( $index, $sock )
{
	global $clients;
	$clients[ $index ][ "sock" ] = $sock;
}

/****************************************************************************************************/
//****** BCS_get_addr
/****************************************************************************************************/
function BCS_get_addr( $index )
{
	global $clients;
	return $clients[ $index ][ "address" ];
}

/****************************************************************************************************/
//****** BCS_set_addr
/****************************************************************************************************/
function BCS_set_addr( $index, $addr )
{
	global $clients;
	$clients[ $index ][ "address" ] = $addr;
}

/****************************************************************************************************/
//****** BCS_get_nick
/****************************************************************************************************/
function BCS_get_nick( $index )
{
	global $clients;
	return $clients[ $index ][ "nick" ];
}

/****************************************************************************************************/
//****** BCS_set_nick
/****************************************************************************************************/
function BCS_set_nick( $index, $nick )
{
	global $clients;
	$clients[ $index ][ "nick" ] = $nick;
}

/****************************************************************************************************/
//****** BCS_get_input
/****************************************************************************************************/
function BCS_get_input( $index )
{
	global $clients;
	return $clients[ $index ][ "input" ];
}

/****************************************************************************************************/
//****** BCS_app_input
/****************************************************************************************************/
function BCS_set_input( $index, $text )
{
	global $clients;
	$clients[ $index ][ "input" ] = $text;
}

/****************************************************************************************************/
//****** BCS_reset_input
/****************************************************************************************************/
function BCS_reset_input( $index )
{
	global $clients;
	$clients[ $index ][ "input" ] = "";
}

?>